// Message handling routines
//
// Call msg to print an error message
// Call msg_set_err_mask to set which errors should be printed
// Call msg_get_err_mask to get current error mask
//
// Copyright 2014 David Gesswein.
// This file is part of MFM disk utilities.
//
// MFM disk utilities is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// MFM disk utilities is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with MFM disk utilities.  If not, see <http://www.gnu.org/licenses/>.
#include <stdio.h>
#include <stdarg.h>
#include <stdint.h>
#include <time.h>

#ifdef CLOCK_MONOTONIC_RAW
#define CLOCK CLOCK_MONOTONIC_RAW
#else
#define CLOCK CLOCK_MONOTONIC
#endif

#include "msg.h"

// Current error mask. Not thread safe for changes
static uint64_t err_mask = 0xffffffff;

// Print an error message.
//
// level: Error level used to determine if message should be printed
// format: Format string for message
//
void msg(uint32_t level, char *format, ...) {
   static int last_progress = 0;
   va_list va;


   if (err_mask & level) {
#if 0
      static double last_msg_time = 0;
      double cur_time;
      struct timespec tv_start;

      clock_gettime(CLOCK, &tv_start);
      cur_time = tv_start.tv_sec + tv_start.tv_nsec / 1e9;
      if (last_msg_time == 0) {
         last_msg_time = cur_time;
      } 
      printf("%.4f ",cur_time - last_msg_time);
      last_msg_time = cur_time;
#endif

      // Overwrite progress message with spaces in case new message is shorter
      if (last_progress && !(level & MSG_PROGRESS)) {
         printf("%79s\r","");
      }
      va_start(va, format);
      vprintf(format, va);
      va_end(va);
      if (level & MSG_PROGRESS) {
         fflush(stdout);
      }
      last_progress = level & MSG_PROGRESS;
   }
}

// Set the error mask and get previous value
//
// mask: New error mask
// return: Previous error mask
uint32_t msg_set_err_mask(uint32_t mask) {
   uint32_t last_mask;

   last_mask = err_mask;
   err_mask = mask;
   return last_mask;
}

// Get the error mask
// return: Current error mask
uint32_t msg_get_err_mask(void) {
   return err_mask;
}
