#define VERSION "0.5"
// Parse the command line.
//
// Call parse_cmdline to parse the command line
// Call parse_print_cmdline to print drive parameter information in command
//   line format
//
// Copyright 2014 David Gesswein.
// This file is part of MFM disk utilities.
//
// MFM disk utilities is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// MFM disk utilities is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with MFM disk utilities.  If not, see <http://www.gnu.org/licenses/>.
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <getopt.h>
#include <stdarg.h>
#include <stdio.h>


#include "msg.h"
#include "crc_ecc.h"
#include "mfm_decoder.h"

#include "parse_cmdline.h"

#define ARRAYSIZE(x) (sizeof(x) / sizeof(x[0]))

void safe_print(char **ptr, int *left, char *format, ...) {
   va_list va;
   int rc;

   va_start(va, format);
   rc = vsnprintf(*ptr, *left, format, va);
   *left -= rc;
   if (left <= 0) {
      msg(MSG_FATAL, "Command line exceeded buffer\n");
      exit(1);
   }
   *ptr += rc;
   va_end(va);
}
// Print command line to give current drive parameter settings
//
// drive_params: Drive parameters
char *parse_print_cmdline(DRIVE_PARAMS *drive_params, int print) {
   int i;
   static char cmdline[2048];
   int cmdleft = sizeof(cmdline)-1;
   char *cmdptr = cmdline;

   safe_print(&cmdptr, &cmdleft, "--sectors %d,%d --heads %d --cylinders %d ",
         drive_params->num_sectors, drive_params->first_sector_number,
         drive_params->num_head, drive_params->num_cyl);
   safe_print(&cmdptr, &cmdleft,
         "--header_crc 0x%llx,0x%llx,%d,%d  --data_crc  0x%llx,0x%llx,%d,%d ",
         drive_params->header_crc.init_value, drive_params->header_crc.poly,
         drive_params->header_crc.length, drive_params->header_crc.ecc_max_span,
         drive_params->data_crc.init_value, drive_params->data_crc.poly,
         drive_params->data_crc.length, drive_params->data_crc.ecc_max_span);
   safe_print(&cmdptr, &cmdleft,
         "--format %s --sector_length %d --retries %d --drive %d",
         controller_names[drive_params->controller], drive_params->sector_size,
         drive_params->retries, drive_params->drive);
   if (!drive_params->buffered_seek) {
      safe_print(&cmdptr, &cmdleft, " --unbuffered_seek");
   }
   if (drive_params->head_3bit) {
      safe_print(&cmdptr, &cmdleft, " --head_3bit");
   }
   if (drive_params->sector_numbers != NULL) {
      safe_print(&cmdptr, &cmdleft, " --interleave ");
      for (i = 0; i < drive_params->num_sectors; i++) {
         if (i == drive_params->num_sectors - 1) {
            safe_print(&cmdptr, &cmdleft, "%d", drive_params->sector_numbers[i]);
         } else {
            safe_print(&cmdptr, &cmdleft, "%d,", drive_params->sector_numbers[i]);
         }
      }
   }
#if 0
   msg(MSG_INFO_SUMMARY," --transitions_file %s --extracted_data_file %s",
         drive_params->transitions_filename, drive_params->extract_filename);
   msg(MSG_INFO_SUMMARY," --emulation_file %s",
         drive_params->emulation_file);
#endif
   if (print) {
      msg(MSG_INFO_SUMMARY, "Command line to read disk:\n%s\n", cmdline);
   }
   return cmdline;
}

// Parse CRC values
//
// arg: CRC value string
// return: CRC information
static CRC_INFO parse_crc(char *arg) {
   CRC_INFO info;
   int i;
   char *str, *tok;
   uint64_t val;

   str = arg;
   memset(&info, 0, sizeof(info));

   for (i = 0; i < 4; i++) {
      tok = strtok(str,",");
      if (tok == NULL) {
         break;
      }
      str = NULL;
      val = strtoull(tok, NULL, 0);
      switch(i) {
      case 0:
         info.init_value = val;
         break;
      case 1:
         info.poly = val;
         break;
      case 2:
         info.length = val;
         break;
      case 3:
         info.ecc_max_span = val;
         break;
      }
   }
   if (i < 3) {
      msg(MSG_FATAL,"Minimum for CRC is initial value, polynomial, and polynomial size\n");
      exit(1);
   }
   return info;
}

// Parse controller value (track/sector header format). The formats are named
// after the controller that wrote the format. Multiple controllers may use
// the same format.
//
// arg: Controller string
// return: Controller number
static int parse_controller(char *arg) {
   int i;
   int controller = -1;

   for (i = 0; controller_names[i] != NULL; i++) {
      if (strcasecmp(controller_names[i], arg) == 0) {
         controller = i;
      }
   }
   if (controller == -1) {
      msg(MSG_FATAL, "Unknown controller %s. Choices are\n",arg);
      for (i = 0; controller_names[i] != NULL; i++) {
         msg(MSG_FATAL,"%s\n",controller_names[i]);
      }
      exit(1);
   }
   return controller;
}

// Parse the interleave information. It may either be an interleave number
// or a comma separated list of sector number.
//
// arg: Interleave information string
// drive_params: Drive parameters
// return: Pointer to list of sector number
static uint8_t *parse_interleave(char *arg, DRIVE_PARAMS *drive_params) {
   // Must be static since we return pointer. Sector numbers for interleave
   static uint8_t sectors[MAX_SECTORS];
   // Sector numbers already used.
   uint8_t used_sectors[MAX_SECTORS];
   int i;
   char *str, *tok;
   int interleave, sectnum;

   str = arg;
   for (i = 0; i < MAX_SECTORS; i++) {
      tok = strtok(str,",");
      if (tok == NULL) {
         break;
      }
      str = NULL;
      sectors[i] = atoi(tok);
   }
   // If one parameter it is interleave value so we calculate the sector
   // numbers for each sector
   if (i == 1) {
      if (drive_params->num_sectors == 0) {
         msg(MSG_FATAL, "Number of sectors must be set before interleave is specified\n");
         exit(1);
      }
      interleave = sectors[0];
      sectnum = 0;
      memset(used_sectors, 0, sizeof(used_sectors));
      for (i = 0; i < drive_params->num_sectors; i++) {
         // Find next unused sector number if already used.
         while (used_sectors[sectnum]) {
            sectnum = (sectnum + 1) % drive_params->num_sectors +
                  drive_params->first_sector_number;
         }
         sectors[i] = sectnum;
         used_sectors[sectnum] = 1;
         sectnum = (sectnum + interleave) % drive_params->num_sectors +
               drive_params->first_sector_number;
      }
   } else {
      if (drive_params->num_sectors != 0 && drive_params->num_sectors != i) {
         msg(MSG_FATAL, "Number of sectors in interleave list doesn't match number of sectors\n");
         exit(1);
      }
   }
   return sectors;
}

// Delete bit n from v shifting higher bits down
#define DELETE_BIT(v, n) (v & ((1 << n)-1)) | (((v & ~((1 << (n+1))-1)) >> 1))

// Main routine for parsing command lines
//
// argc, argv: Main argc, argv
// drive_parameters: Drive parameters where most of the parsed values are stored
// analyze: 1 if analyze was specified on command line
// delete_options: Options to delete from list of valid options
void parse_cmdline(int argc, char *argv[], DRIVE_PARAMS *drive_params, int *analyze,
      char *delete_options)
{
   // Minimum to generate extract file (sector data)
   int min_read_opts = 0x7f;
   // Minimum to generation emulation file (MFM clock & data)
   int min_read_transitions_opts = 0x46;
   // If you change this fix parse_print_cmdline
   struct option long_options[] = {
         {"sectors", 1, NULL, 's'},
         {"heads", 1, NULL, 'h'},
         {"cylinders", 1, NULL, 'c'},
         {"header_crc", 1, NULL, 'g'},
         {"data_crc", 1, NULL, 'd'},
         {"format", 1, NULL, 'f'},
         {"drive", 1, NULL, 'j'},
         {"sector_length", 1, NULL, 'l'},
         {"unbuffered_seek", 0, NULL, 'u'},
         {"interleave", 1, NULL, 'i'},
         {"head_3bit", 0, NULL, '3'},
         {"retries", 1, NULL, 'r'},
         {"analyze", 0, NULL, 'a'},
         {"quiet", 1, NULL, 'q'},
         {"transitions_file", 1, NULL, 't'},
         {"extracted_data_file", 1, NULL, 'e'},
         {"emulation_file", 1, NULL, 'm'},
         {"version", 0, NULL, 'v'},
         {NULL, 0, NULL, 0}
   };
   char short_options[] = "s:h:c:g:d:f:l:ui:3r:aq:t:e:j:v";
   int rc;
   // Loop counters
   int i,j;
   int options_index;
   // Bit vector of which options were specified
   uint32_t opt_mask = 0;
   char *tok;

   // Enable all errors other than debug
   msg_set_err_mask(~0 ^ MSG_DEBUG);
   // Options above are superset for all the programs to ensure options stay consistent
   // Each caller can specify options to delete.
   for (i = 0; delete_options != NULL && i < strlen(delete_options); i++) {
      tok = strchr(short_options,delete_options[i]);
      if (tok == NULL) {
         msg(MSG_FATAL, "Didn't find option %c in list %s\n",delete_options[i], short_options);
         exit(1);
      } else {
         // Delete option and possible following : from short options
         if (*(tok+1) == ':') {
            memmove(tok, tok+2, strlen(tok+2)+1);
         } else {
            memmove(tok, tok+1, strlen(tok+1)+1);
         }
      }
      // And delete long option
      for (j = 0; long_options[j].name != NULL; j++) {
         if (long_options[j].val == delete_options[i]) {
            min_read_transitions_opts =
                  DELETE_BIT(min_read_transitions_opts, j);
            min_read_opts = DELETE_BIT(min_read_opts, j);
            memmove(&long_options[j], &long_options[j+1], sizeof(long_options[0]) *
                  (ARRAYSIZE(long_options) - j - 1));
            break;
         }
      }
   }
   memset(drive_params, 0, sizeof(*drive_params));
   // Set defaults
   drive_params->emu_fd = -1;
   drive_params->tran_fd = -1;
   drive_params->buffered_seek = 1;
   drive_params->retries = 50;
   drive_params->sector_size = 512;
   drive_params->emulation_output = 1; // This will be adjusted by caller
   *analyze = 0;
   // Handle the options. The long options are converted to the short
   // option name for the switch by getopt_long.
   options_index = -1;
   while ((rc = getopt_long(argc, argv, short_options, long_options,
         &options_index)) != -1) {
      // Short options don't set options_index so look it up
      if (options_index == -1) {
         for (i = 0; i < ARRAYSIZE(long_options); i++) {
            if (rc == long_options[i].val) {
               options_index = i;
               break;
            }
         }
         if (options_index == -1) {
            //msg(MSG_FATAL, "Error parsing option %c\n",rc);
            msg(MSG_FATAL,"Valid options:\n");
            for (i = 0; long_options[i].name != NULL; i++) {
              msg(MSG_FATAL, "%c %s\n", long_options[i].val, 
                 long_options[i].name);
            }
            exit(1);
         }
      }
      opt_mask |= 1 << options_index;
      switch(rc) {
      case 's':
         tok = strtok(optarg,",");
         drive_params->num_sectors = atoi(tok);
         tok = strtok(NULL,",");
         if (tok != NULL) {
            drive_params->first_sector_number = atoi(tok);
         }
         if (drive_params->num_sectors > MAX_SECTORS) {
            msg(MSG_FATAL,"Maximum number of sectors %d:\n", MAX_SECTORS);
         }
         break;
      case 'h':
         drive_params->num_head = atoi(optarg);
         if (drive_params->num_head > MAX_HEAD) {
            msg(MSG_FATAL,"Maximum number of heads %d:\n", MAX_HEAD);
         }
         break;
      case 'c':
         drive_params->num_cyl = atoi(optarg);
         break;
      case 'g':
         drive_params->header_crc = parse_crc(optarg);
         break;
      case 'd':
         drive_params->data_crc = parse_crc(optarg);
         break;
      case 'u':
         drive_params->buffered_seek = 0;
         break;
      case 'i':
         drive_params->sector_numbers = parse_interleave(optarg, drive_params);
         break;
      case '3':
         drive_params->head_3bit = 1;
         break;
      case 'f':
         drive_params->controller = parse_controller(optarg);
         break;
      case 'l':
         drive_params->sector_size = atoi(optarg);
         if (drive_params->sector_size > MAX_SECTOR_SIZE) {
            msg(MSG_FATAL,"Maximum sector size %d:\n", MAX_SECTOR_SIZE);
         }
         break;
      case 'r':
         drive_params->retries = atoi(optarg);
         break;
      case 'a':
         *analyze = 1;
         break;
      case 't':
         drive_params->transitions_filename = optarg;
         break;
      case 'e':
         drive_params->extract_filename = optarg;
         break;
      case 'm':
         drive_params->emulation_filename = optarg;
         break;
      case 'j':
         drive_params->drive = atoi(optarg);
         break;
      case 'q':
         msg_set_err_mask(~strtoul(optarg, NULL, 0));
         break;
      case 'v':
         msg(MSG_INFO_SUMMARY,"Version %s\n",VERSION);
         break;
      case '?':
         msg(MSG_FATAL, "Didn't process argument %c\n", rc);
         exit(1);
         break;
      default:
         msg(MSG_FATAL, "Didn't process argument %c\n", rc);
         exit(1);
      }
      options_index = -1;
   }
   if ((drive_params->extract_filename != NULL)
         && !*analyze && (opt_mask & min_read_opts) != min_read_opts) {
      msg(MSG_FATAL, "Generating extract file without analyze requires options:");
      for (i = 0; i < 32; i++) {
         int bit = (1 << i);
         if (!(opt_mask & bit) && (min_read_opts & bit)) {
            msg(MSG_FATAL, " %s", long_options[i].name);
         }
      }
      msg(MSG_FATAL, "\n");
      exit(1);
   } else if ((drive_params->transitions_filename != NULL || drive_params->emulation_filename != NULL)
         && !*analyze && (opt_mask & min_read_transitions_opts) !=
               min_read_transitions_opts) {
      msg(MSG_FATAL, "Generation of transition or emulation file without analyze requires options:\n");
      for (i = 0; i < 32; i++) {
         int bit = (1 << i);
         if (!(opt_mask & bit) && (min_read_transitions_opts & bit)) {
            msg(MSG_FATAL, " %s", long_options[i].name);
         }
      }
      msg(MSG_FATAL, "\n");
      exit(1);
   }
   //parse_print_cmdline(drive_params);
}
